/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * Id : ${file_name} 154 ${date}  rbarraza
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.binding.filetransferbc.listeners;

import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.Timer;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.activation.DataHandler;
import javax.activation.FileDataSource;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.NormalizedMessage;

import org.apache.commons.io.FileUtils;
import org.objectweb.petals.binding.filetransferbc.FileTransferBCException;
import org.objectweb.petals.binding.filetransferbc.IOUtils;
import org.objectweb.petals.common.descriptor.Consumes;
import org.objectweb.petals.common.descriptor.Extensions;
import org.objectweb.petals.common.util.PeriodicTask;
import org.objectweb.petals.common.util.StringHelper;
import org.objectweb.petals.component.common.AbstractComponent;
import org.objectweb.petals.component.common.bc.AbstractBindingComponent;
import org.objectweb.petals.component.common.su.SimpleServiceUnitManager;
import org.objectweb.petals.component.common.util.PetalsExtensionsUtil;
import org.objectweb.petals.component.common.util.SourceHelper;

/**
 * The task which scans the file transfer working directories. It builds JBI
 * message with files as attachments and send them to the defined endpoint.
 * 
 * @version $Rev: 250 $Date: {date}
 * @since Petals 1.0
 * @author Christophe HAMERLING - eBM WebSourcing
 * 
 */
public class FileTransferBCListener extends PeriodicTask {

    private String address;

    private FileTransferFilter filter;

    private Logger logger;

    private AbstractBindingComponent component;

    private String pattern;

    private File outputDir;

    private File wasteDir;

    private boolean attachmentMode;

    private String rootelement;

    /**
     * Creates a new instance of {@link FileTransferBCListener}
     * 
     * @param logger
     * @param channel
     * @param context
     * @param jbiListener
     * @param mapDirConfig
     * @param dir
     */
    public FileTransferBCListener(Logger logger,
            AbstractBindingComponent component, String address,
            long pollingPeriod) {
        super(0, pollingPeriod);
        this.component = component;
        this.logger = logger;
        this.address = address;
    }

    public void run() {
        File inDir = new File(address);
        if (inDir.exists() && inDir.listFiles().length > 0) {
            handleFiles(inDir);
        }
    }

    /**
     * Init data
     * 
     */
    public void init() throws FileTransferBCException {

        // Get required datas
        SimpleServiceUnitManager suManager = (SimpleServiceUnitManager) component
                .getServiceUnitManager();
        Consumes consumes = suManager.getConsumesFromAddress(address);
        Extensions extensions = consumes.getExtensions();

        // get output directory
        String out = retrieveOutputDir(extensions);

        // get the file filter
        filter = new FileTransferFilter(retrieveFileNameFilter(extensions));

        // create the output direcotry if it not exists
        outputDir = new File(out);
        if (!outputDir.exists()) {
            outputDir.mkdirs();
        }

        // create the waste dir used to save transfered files
        wasteDir = new File(outputDir, "waste");
        if (!wasteDir.exists()) {
            wasteDir.mkdirs();
        }

        // get MEP pattern
        this.pattern = PetalsExtensionsUtil.extractValueFromKeyValueExtension(
                consumes.getExtensions(), "pattern");

        // set attachment mode
        this.attachmentMode = Boolean.parseBoolean(PetalsExtensionsUtil
                .extractValueFromKeyValueExtension(extensions, "attachments"));

        // Set rootelement
        this.rootelement = PetalsExtensionsUtil
                .extractValueFromKeyValueExtension(extensions, "rootelement");
    }

    /**
     * Retrieve the output directory from the consumes bean. If the output
     * directory has not been set, set the default directory to the component
     * installation root.
     * 
     */
    protected String retrieveOutputDir(Extensions extensions) {
        String tmp = PetalsExtensionsUtil.extractValueFromKeyValueExtension(
                extensions, "output-dir");

        // output dir has not been set, get JBI output directory
        if (tmp == null) {
            tmp = getDefaultDirectoryName();
        }

        return tmp;
    }

    /**
     * Get the default directory name
     * 
     * @return
     */
    protected String getDefaultDirectoryName() {
        String baseDir = component.getContext().getInstallRoot();
        StringBuffer sb = new StringBuffer(baseDir.substring(0, baseDir
                .length()
                - "install".length()));
        sb.append("work");
        sb.append(File.separator);
        return sb.toString();
    }

    /**
     * Handle an InOnly message
     * 
     * @param body
     * @param attachments
     * @return
     * @throws FileTransferBCException
     */
    public boolean handleInOnlyMessage(String body, Set<DataHandler> attachments)
            throws FileTransferBCException {
        boolean result = false;

        try {

            Map<String, DataHandler> attachmentsMap = createAttachmentMap(attachments);

            MessageExchange exchange = component.sendMessage(address,
                    SourceHelper.createSource(body), attachmentsMap);
            if (!ExchangeStatus.DONE.equals(exchange.getStatus())) {
                logger.log(Level.SEVERE,
                        "A file has not been correctly handled");
            }

        } catch (Exception ex) {
            throw new FileTransferBCException(ex);
        }

        return result;
    }

    private Map<String, DataHandler> createAttachmentMap(
            Set<DataHandler> attachments) {
        Map<String, DataHandler> attachmentsMap = new HashMap<String, DataHandler>();
        // fill attachments
        for (DataHandler dh : attachments) {
            attachmentsMap.put(dh.getName(), dh);
        }
        return attachmentsMap;
    }

    /**
     * Handle an in opt out message. Send the files through JBI and wait for a
     * response. If the response contains files, the files are written to disk.
     * 
     * @param body
     * @param attachments
     * @throws FileTransferBCException
     */
    public void handleInOptOutMessage(String body, Set<DataHandler> attachments)
            throws FileTransferBCException {
        String response = "";
        String fileName = "";
        DeliveryChannel channel = component.getChannel();

        try {
            Map<String, DataHandler> attachmentsMap = createAttachmentMap(attachments);

            MessageExchange exchange = component.sendMessage(address,
                    SourceHelper.createSource(body), attachmentsMap);

            // Verify if request was success,if not, handle the error message
            if (!ExchangeStatus.DONE.equals(exchange.getStatus())) {
                if (exchange.getStatus().equals(ExchangeStatus.ERROR)) {
                    fileName = "error";
                    response = "A file has not been correctly handled";
                } else {
                    // if there wasn't errors , binding component handle out
                    // reponse
                    // or fault
                    NormalizedMessage nmOut = exchange.getMessage("out");
                    if (nmOut != null) {
                        fileName = "content";
                        response = SourceHelper
                                .createString(nmOut.getContent());
                        IOUtils.writeAttachmentsToFiles(nmOut, outputDir);
                    } else {
                        Fault fault = exchange.getFault();
                        if (fault != null) {
                            fileName = "fault";
                            response = SourceHelper.createString(fault
                                    .getContent());
                            IOUtils.writeAttachmentsToFiles(fault, outputDir);
                        }
                    }
                }
                IOUtils.writeStringToFile(response, outputDir, fileName);
                // Send status DONE to complete the exchange
                exchange.setStatus(ExchangeStatus.DONE);
                channel.send(exchange);
            }

        } catch (Exception ex) {
            throw new FileTransferBCException(ex);
        }
    }

    /**
     * Handle a robust in only message. Send the files through JBI and wait for
     * a response. If the response contains files, the files are written to
     * disk.
     * 
     * @param body
     * @param attachments
     * @throws FileTransferBCException
     */
    public void handleRobustInOnlyMessage(String body,
            Set<DataHandler> attachments) throws FileTransferBCException {
        String response = "";
        String fileName = "";
        DeliveryChannel channel = component.getChannel();

        try {
            Map<String, DataHandler> attachmentsMap = createAttachmentMap(attachments);

            MessageExchange exchange = component.sendMessage(address,
                    SourceHelper.createSource(body), attachmentsMap);

            // Verify if request was success,if not, handle the error message
            if (!ExchangeStatus.DONE.equals(exchange.getStatus())) {
                if (exchange.getStatus().equals(ExchangeStatus.ERROR)) {
                    fileName = "error";
                    response = "A file has not been correctly handled";
                } else {
                    // if there wasn't errors , binding component handle fault
                    Fault fault = exchange.getFault();
                    if (fault != null) {
                        fileName = "fault";
                        response = SourceHelper
                                .createString(fault.getContent());
                        IOUtils.writeAttachmentsToFiles(fault, outputDir);
                    }
                }
                IOUtils.writeStringToFile(response, outputDir, fileName);

                // Send status DONE to complete the exchange
                exchange.setStatus(ExchangeStatus.DONE);
                channel.send(exchange);
            }

        } catch (Exception ex) {
            throw new FileTransferBCException(ex);
        }
    }

    /**
     * Handle an in opt out message. Send the files through JBI and wait for a
     * response. If the response contains files, the files are written to disk.
     * 
     * @param body
     * @param attachments
     * @throws FileTransferBCException
     */
    public void handleInOutMessage(String body, Set<DataHandler> attachments)
            throws FileTransferBCException {
        String response = "";
        String fileName = "";
        DeliveryChannel channel = component.getChannel();

        try {
            Map<String, DataHandler> attachmentsMap = createAttachmentMap(attachments);

            MessageExchange exchange = component.sendMessage(address,
                    SourceHelper.createSource(body), attachmentsMap);

            // if there wasn't errors , binding component handle out reponse
            // or fault
            NormalizedMessage nmOut = exchange.getMessage("out");
            if (nmOut != null) {
                fileName = "content";
                response = SourceHelper.createString(nmOut.getContent());
                IOUtils.writeAttachmentsToFiles(nmOut, outputDir);
            } else {
                Fault fault = exchange.getFault();
                if (fault != null) {
                    fileName = "fault";
                    response = SourceHelper.createString(fault.getContent());
                    IOUtils.writeAttachmentsToFiles(fault, outputDir);
                }
            }

            IOUtils.writeStringToFile(response, outputDir, fileName);

            // Send status DONE to complete the exchange
            exchange.setStatus(ExchangeStatus.DONE);
            channel.send(exchange);

        } catch (Exception ex) {
            throw new FileTransferBCException(ex);
        }
    }

    /**
     * Handle the files of the directory
     * 
     * @param srcDir
     * @param inOut
     */
    private void handleFiles(File srcDir) {

        File[] srcFiles = srcDir.listFiles(filter);

        // Create body and attachements
        try {
            String body = "";
            Set<DataHandler> attachments = new HashSet<DataHandler>();
            if (attachmentMode) {
                body = "<?xml version='1.0' encoding='UTF-8'?><info>attached file(s)</info>";
                for (File src : srcFiles) {
                    if (!src.isDirectory()) {
                        DataHandler dhFile = new DataHandler(
                                new FileDataSource(src));
                        attachments.add(dhFile);
                    }
                }
                sendMessage(body, attachments);
            } else {
                for (File srcFile : srcFiles) {
                    try {
                        body = FileUtils.readFileToString(srcFile, null);
                        if (!StringHelper.isNullOrEmpty(rootelement)) {
                            body = new StringBuffer("<").append(rootelement)
                                    .append(">").append(body).append("</")
                                    .append(rootelement).append(">").toString();
                        }
                        sendMessage(body, attachments);
                    } catch (IOException e) {
                        this.logger.log(Level.WARNING, e.getMessage());
                    }
                }
            }
        } finally {
            moveFilesToWaste(srcFiles);
        }

    }

    /**
     * Send files through JBI. InOnly pattern is used by default.
     * 
     * @param inOut
     * @param body
     * @param attachments
     */
    private void sendMessage(String body, Set<DataHandler> attachments) {
        try {
            if (AbstractComponent.IN_OUT.equalsIgnoreCase(pattern)) {
                handleInOutMessage(body, attachments);
            } else if (AbstractComponent.IN_OPTIONAL_OUT
                    .equalsIgnoreCase(pattern)) {
                handleInOptOutMessage(body, attachments);
            } else if (AbstractComponent.ROBUST_IN_ONLY
                    .equalsIgnoreCase(pattern)) {
                handleRobustInOnlyMessage(body, attachments);
            } else {
                handleInOnlyMessage(body, attachments);
            }
        } catch (FileTransferBCException e) {
            this.logger.log(Level.SEVERE, e.getMessage(), e);
        }
    }

    /**
     * Move the files to the waste directory.
     * 
     * @param files
     *            the files to move
     * @param parentDir
     *            the parent directory (waste directory is included)
     */
    private void moveFilesToWaste(File[] files) {
        for (File file : files) {
            file.renameTo(new File(wasteDir, file.getName()));
        }
    }

    /**
     * Get the file filter value from the extensions. This filter will be used
     * to filter the files to handle in the polling directory.
     * 
     * @return the file filter or empty string if not found
     */
    protected String retrieveFileNameFilter(Extensions extensions) {
        String filter = "";
        String tmp = PetalsExtensionsUtil.extractValueFromKeyValueExtension(
                extensions, "filename-filter");
        if (tmp != null)
            filter = tmp;

        return filter;
    }

    /**
     * Get the timer for this task
     * 
     * @return
     */
    public Timer getTimer() {
        return this.timer;
    }

    /**
     * A file name filter implementation. Accept will return true if the file
     * name contains the filter value. If filter is an empty string, the filter
     * acts as nothing, all the files are accepted.
     * 
     * @author Christophe HAMERLING - eBM WebSourcing
     * 
     */
    protected class FileTransferFilter implements FilenameFilter {
        /**
         * The filter to apply
         */
        private String filter;

        /**
         * Creates a new instance of FileTransferFilter
         * 
         * @param filter
         */
        protected FileTransferFilter(String filter) {
            if (filter == null)
                filter = "";
            this.filter = filter;
        }

        /*
         * (non-Javadoc)
         * 
         * @see java.io.FilenameFilter#accept(java.io.File, java.lang.String)
         */
        public boolean accept(File dir, String name) {
            return (name.indexOf(filter) != -1);
        }
    }

}
